
#ifndef PLS_DATAFLOW_INTERNAL_GRAPH_H_
#define PLS_DATAFLOW_INTERNAL_GRAPH_H_

#include <atomic>
#include <tuple>
#include <iostream>

#include "node.h"
#include "in_port.h"
#include "out_port.h"
#include "inputs.h"
#include "outputs.h"

namespace pls {
namespace dataflow {
namespace internal {

template<typename INS, typename OUTS>
class graph {};

template<typename I0, typename ...I, typename O0, typename ...O>
class graph<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>> : public node {
 private:
  // Our own type
  using self_type = graph<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>>;

  // Callbacks for input/output ports
  struct in_port_cb {
    self_type *self_;
    template<int POS, typename T>
    void token_pushed(token<T> token) {
      self_->in_port_token_pushed<POS, T>(token);
    }
  };
  struct out_port_cb {
    self_type *self_;
    template<int POS, typename T>
    void token_pushed(token<T> token) {
      self_->out_port_token_pushed<POS, T>(token);
    }
  };

  // Input-Output port types (external)
  using multi_in_port_type = multi_in_port<in_port_cb, 0, I0, I...>;
  using multi_out_port_type = multi_out_port<O0, O...>;

  // Input-Output port types (internal)
  using internal_in_port_type = multi_out_port<I0, I...>;
  using internal_out_port_type = multi_in_port<out_port_cb, 0, O0, O...>;

  // Input-Output value tuples
  using value_input_tuple = std::tuple<I0, I...>;
  using input_tuple = std::tuple<token < I0>, token<I>...>;
  using output_tuple = std::tuple<token < O0>, token<O>...>;
  static constexpr int num_in_ports = std::tuple_size<input_tuple>();
  static constexpr int num_out_ports = std::tuple_size<output_tuple>();

  // Memory used for ONE active invocation of the dataflow-graph
  struct invocation_memory {
    std::atomic<int> inputs_missing_;
    input_tuple input_buffer_;
  };

 public:
  template<int POS>
  using in_port_at = typename multi_in_port_type::template in_port_type_at<POS>;
  template<int POS>
  using out_port_at = typename multi_out_port_type::template out_port_type_at<POS>;

  template<int POS>
  in_port_at<POS> &in_port() {
    return in_port_.template get<POS>();
  }
  template<int POS>
  out_port_at<POS> &out_port() {
    return out_port_.template get<POS>();
  }

  template<int POS, typename T>
  void in_port_token_pushed(token<T> token) {
    auto invocation = get_invocation<invocation_memory>(token);

    // TODO: Invoke a new instance of our graph (recursive call) when all inputs are here
    std::cout << "Token pushed at " << POS << " with value " << token.value() << std::endl;
    auto remaining_inputs = --(invocation->inputs_missing_);
    if (remaining_inputs == 0) {
      std::cout << "All inputs there!" << std::endl;
    }
  }

  template<int POS>
  using internal_in_port_at = typename internal_in_port_type::template out_port_type_at<POS>;
  template<int POS>
  using internal_out_port_at = typename internal_out_port_type::template in_port_type_at<POS>;

  template<int POS>
  internal_in_port_at<POS> &input() {
    return internal_in_port_.template get<POS>();
  }
  template<int POS>
  internal_out_port_at<POS> &output() {
    return internal_out_port_.template get<POS>();
  }

  template<int POS, typename T>
  void out_port_token_pushed(token<T> token) {
    auto invocation = get_invocation<invocation_memory>(token);

    std::cout << "Output produced at port " << POS << " with value " << token.value() << std::endl;
    auto remaining_inputs = --(invocation->inputs_missing_);
    if (remaining_inputs == 0) {
      std::cout << "All inputs there!" << std::endl;
    }
  }

  graph() : in_port_cb_{this},
            out_port_cb_{this},
            in_port_{this, &in_port_cb_},
            out_port_{},
            internal_in_port_{},
            internal_out_port_{this, &out_port_cb_} {
  }

  /////////////////////////////////////////////////////////////////
  // Graph building
  /////////////////////////////////////////////////////////////////
  void build() {
    state_ = building;
    PLS_ASSERT(is_internal_fully_connected(), "Must fully connect all inputs/outputs inside a dataflow graph!")

    num_nodes_ = 0;
    for (int i = 0; i < num_in_ports; i++) {
      build_recursive(internal_in_port_.next_node_at(i));
    }

    state_ = built;
  }

  node *start_{nullptr};
  node *current_{nullptr};

  void build_recursive(node *node) {
    if (node->state_ != fresh) {
      return; // Already visited
    }
    node->state_ = building;
    PLS_ASSERT(node->is_fully_connected(), "Must fully connect dataflow graph nodes!")

    add_node(node);
    for (int i = 0; i < node->num_successors(); i++) {
      build_recursive(node->successor_at(i));
    }

    node->state_ = built;
  }

  void add_node(node *new_node) {
    new_node->memory_index_ = num_nodes_++;

    if (start_ == nullptr) {
      start_ = new_node;
      current_ = new_node;
    } else {
      current_->direct_successor_ = new_node;
      current_ = new_node;
    }
  }

  // TODO: Clean up + move to async task
  template<int N, typename ...IT>
  struct feed_inputs {
    feed_inputs(internal_in_port_type &, value_input_tuple &, invocation_info &) {}
    void run() {}
  };
  template<int N, typename IT1, typename ...IT>
  struct feed_inputs<N, IT1, IT...> {
    internal_in_port_type &inputs_;
    value_input_tuple &input_values_;
    invocation_info &invocation_;

    feed_inputs(internal_in_port_type &inputs,
                value_input_tuple &input_values,
                invocation_info &invocation) : inputs_{inputs},
                                               input_values_{input_values},
                                               invocation_{invocation} {}

    void run() {
      inputs_.template get<N>().push_token(token<IT1>{std::get<N>(input_values_), invocation_});
      feed_inputs<N + 1, IT...>{inputs_, input_values_, invocation_}.run();
    }
  };

  void run(value_input_tuple input) {
    // TODO: clearly move this onto the task stack instead fo malloc (without any free...)
    void **buffers = reinterpret_cast<void **>(malloc(num_nodes_ * sizeof(void *)));
    node *iterator = start_;
    for (int i = 0; i < num_nodes_; i++) {
      auto required_size = iterator->instance_buffer_size();
      buffers[i] = malloc(required_size);
      iterator->init_instance_buffer(buffers[i]);

      iterator = iterator->direct_successor_;
    }

    invocation_info invocation{buffers};
    feed_inputs<0, I0, I...>{internal_in_port_, input, invocation}.run();
  }

  //////////////////////////////////////////////////////////////////
  // Overrides for generic node functionality (building the graph)
  //////////////////////////////////////////////////////////////////
  int num_successors() const override {
    return num_out_ports;
  }
  node *successor_at(int pos) const override {
    return out_port_.next_node_at(pos);
  }

  bool is_fully_connected() const override {
    return in_port_.fully_connected() && out_port_.fully_connected();
  }

  bool is_internal_fully_connected() const {
    return internal_in_port_.fully_connected() && internal_out_port_.fully_connected();
  }

  int instance_buffer_size() const override {
    return sizeof(invocation_memory);
  }
  void init_instance_buffer(void *memory) const override {
    auto invocation = new(memory) invocation_memory{};
    invocation->inputs_missing_ = num_in_ports;
  };

 private:
  // Input/Output ports
  in_port_cb in_port_cb_;
  out_port_cb out_port_cb_;

  multi_in_port_type in_port_;
  multi_out_port_type out_port_;

  internal_in_port_type internal_in_port_;
  internal_out_port_type internal_out_port_;

  int num_nodes_{0};
};

}
}
}

#endif //PLS_DATAFLOW_INTERNAL_GRAPH_H_
